//Request manager takes in the latest packet and processes it as a request
//This means it separates the packet data into the proper components
//It also sends an instruction to the serial transmitter on what to return to the sender
//It also checks to see if the available module is available, and if not waits on it
module RequestHandler(
							clock,
							packet_in,
							packet_crc_in,
							packet_available,
							oxstart,
							oystart,
							oxend_width,
							oyend_height,
							ocolor,
							oinstr,
							oextra1,
							oextra2,
							reset,
							grab_out,
							nodeWanted,
							nodeWanted_isAvailable,
							command_out,
							sync,
							transmit_clock,
							isTransmitting,
							my_state,
							packetID,
							frame_read,
							frame_write,
							bug_check,
							SW,
							IRAM_out,
							drawMode
							);

output reg [2:0] drawMode;
input wire [17:0] SW;
output wire [79:0] IRAM_out;
assign IRAM_out = IRAM_instrOut;

input wire clock;
input wire [79:0] packet_in;
input wire packet_crc_in;
input wire packet_available;
input wire reset;
input wire nodeWanted_isAvailable;
input wire sync;
input wire isTransmitting;
input wire [7:0] packetID;

output reg [5:0] oinstr;
output reg [11:0] ocolor;
output reg [9:0] oxstart, oxend_width;
output reg [8:0] oystart, oyend_height;
output reg [7:0] oextra1, oextra2;
output reg grab_out;
output reg [3:0] nodeWanted;
output reg command_out;
output reg transmit_clock;

output wire [3:0] my_state;
assign my_state = {3'b0,writeToIRAM};

output wire  frame_read, frame_write;
reg readFrame, writeFrame;
assign frame_read = readFrame;
assign frame_write = writeFrame;

reg [3:0] state;
reg [7:0] bugCounter;
output wire [7:0] bug_check = lastPacketID;

parameter 	[3:0] 	waitForPacket = 4'd0,
					grabPacket = 4'd1,
					processPacketInformation = 4'd2,
					waitForAvailableNode = 4'd3,
					sendCommand = 4'd4,
					waitForNodeActive = 4'd5,
					waitForNodeComplete = 4'd6,
					sendCRCResultBack = 4'd7,
					packetNotHere = 4'd8;
					
reg [7:0] lastPacketID;


reg writeToIRAM;
reg [79:0] instrToWrite;
reg [11:0] PC_write;
wire [11:0] PC;
wire [79:0] IRAM_instrOut;
wire isDraw;
/*SimpleIRAM(
			.clock(clock), 
			.write(writeToIRAM), 
			.instr_in(instrToWrite), 
			.pc(SW[17:6]), 
			.pc_write(PC_write),
			.instr_out(IRAM_instrOut), 
			.isDraw(isDraw), 
			.reset(reset)
			);
*/
assign PC = SW[3] == 1'b1 ? SW[15:4] : ((instructionType == memory) ? PC_memory : packet_in[23:12]);
ComplexIRAM cIRAM1(
	.address(PC),
	.clock(clock),
	.data(packet_in),
	.wren(writeToIRAM),
	.q(IRAM_instrOut)
	);
	
reg writeIRAM_clock;
	
Processor proc1(
                    .clock_50(clock),
                    .draw_complete(drawComplete),
                    .reset(reset),
                    .enable(processorEnabled),
                    .instruction(IRAM_instrOut),
                    .pc(PC_memory),
                    .isDrawInstruction(isDrawInstruction)
                    );
                    
reg [1:0] instructionType;
parameter 	direct = 2'b00,
			memory = 2'b01,
			download = 2'b10;
wire [11:0] PC_memory;
wire processorEnabled;
assign processorEnabled = (instructionType == memory);
reg drawComplete;
wire isDrawInstruction;

reg [2:0] processor_state;
parameter waitInstruction = 3'd0,
			parseInstruction = 3'd1,
			selectNode = 3'd2,
			issueCommand = 3'd3,
			waitForNodeToStart = 3'd4,
			waitForNodeToFinish = 3'd5,
			cleanUp = 3'd6;
			
reg [2:0] download_state;
parameter waitForDownloadPacket = 3'd0,
			writePacket = 3'd1,
			sendACK = 3'd2,
			waitTwo = 3'd3,
			enter = 3'd4;

always @(posedge clock) begin

	if(reset) begin
		state <= waitForPacket;
		oinstr <= 6'b0;
		ocolor <= 12'b0;
		oxstart <= 10'b0;
		oxend_width <= 10'b0;
		oystart <= 9'b0;
		oyend_height <= 9'b0;
		oextra1 <= 8'b0;
		oextra2 <= 8'b0;
		grab_out <= 1'b0;
		nodeWanted <= 4'b0;
		command_out <= 1'b0;
		transmit_clock <= 1'b0;
		lastPacketID <= 8'd255;
		readFrame <= 1'b0;
		writeFrame <= 1'b1;
		writeToIRAM <= 1'b0;
		instructionType = direct;
		drawComplete <= 1;
		processor_state <= waitInstruction;
		drawMode <= 0;
		download_state <= waitForDownloadPacket;
		writeIRAM_clock <= 1'b0;
	end
	else begin
	//if(sync | 1'b1) begin
		if(instructionType == direct) begin
		case(state)
			waitForPacket: begin
				if(packet_available) begin
					state <= grabPacket;
					grab_out <= 1'b0;
				end
				transmit_clock <= 1'b0;
			end
			
			grabPacket: begin
				grab_out <= 1'b1;
				
				//if(~packet_available | 1'b1) begin //wait for grab
				//if((lastPacketID != packetID) ) begin
					state <= processPacketInformation;
					lastPacketID <= packetID;
				//end
				/*
				else begin
					bugCounter <= bugCounter + 1;
				end*/
				
				/*if(~packet_available) begin
					state <= processPacketInformation;
				end*/
				//if(~packet_available) begin //for testing, meaning we grabbed it
					//state <= processPacketInformation;
					//lastPacketID <= packetID;
				//end
				
				if(packet_in[79:74] == 6'd8) begin
					transmit_clock <= 1'b1;
					instructionType = download;
					state <= waitForPacket;
					download_state <= enter;
				end
				
			end
			
			processPacketInformation: begin
				oinstr <= packet_in[79:74];
				ocolor <= packet_in[73:62];
				oxstart <= packet_in[61:52]; //x_start
				oystart <= packet_in[51:43]; //y_start
				oxend_width <= packet_in[42:33]; //x_end/width
				oyend_height <= packet_in[32:24]; //y_end/height
				oextra1 <= packet_in[23:16]; //extra_1
				oextra2 <= packet_in[15:8]; //extra_2
				grab_out <= 1'b0;
				bugCounter <= 0;
				/*if(packet_in[11:8] == 4'hF) begin
					//PC_write <= packet_in[23:12];
					//instrToWrite <= packet_in;
					state <= packetNotHere;
					transmit_clock <= 1'b1;
					writeToIRAM <= 1'b1;
				end
				else begin*/
				if(packet_in[79:74] == 6'd6) begin
					state <= packetNotHere;
					readFrame <= ~readFrame;
					writeFrame <= ~writeFrame;
					transmit_clock <= 1'b1;
				end
				else begin
					state <= waitForAvailableNode;
				end
				/*case(packet_in[11:8])
					4'hF: begin
						state <= packetNotHere;
						transmit_clock <= 1'b1;
						writeToIRAM <= 1'b1;
					end
					default: begin
						case(packet_in[79:74])
							6'd6: begin //FRAME SWITCH
								state <= packetNotHere;
								readFrame <= ~readFrame;
								writeFrame <= ~writeFrame;
								transmit_clock <= 1'b1;
							end
							default: begin
								state <= waitForAvailableNode;
							end
						endcase
					end
				endcase*/
				/*
					if((packet_in[79:74] != 6'd6) && (packet_in[11:8] != 4'hF)) begin
						state <= waitForAvailableNode;
					end
					else begin
						if((packet_in[79:74] == 6'd6) && (packet_in[11:8] != 4'hF)) begin
							state <= packetNotHere;
							readFrame <= ~readFrame;
							writeFrame <= ~writeFrame;
							transmit_clock <= 1'b1;
						end
						else begin
							state <= packetNotHere;
							transmit_clock <= 1'b1;
							writeToIRAM <= 1'b1;
						end
					
					end*/
				//end
				/*
				if(packet_in[79:8] == {72{1'b1}}) begin
					//state <= packetNotHere;
					transmit_clock <= 1'b1;
					instructionType <= memory;
					processor_state <= waitInstruction;
					//PC_memory <= 0;
				end
				else if(packet_in[79:74] == 6'd7 && ~(packet_in[11:8] == 4'd15)) begin
					state <= packetNotHere;
					transmit_clock <= 1'b1;
					drawMode <= packet_in[23:21];
				end
				else if(packet_in[79:74] == 6'd6 && ~(packet_in[11:8] == 4'd15)) begin//frame update
					state <= packetNotHere;
					readFrame <= ~readFrame;
					writeFrame <= ~writeFrame;
					transmit_clock <= 1'b1;
				end
				else if(packet_in[11:8] == 4'd15) begin //writing to IRAM
					PC_write <= packet_in[23:12];
					instrToWrite <= packet_in;
					state <= packetNotHere;
					transmit_clock <= 1'b1;
					writeToIRAM <= 1'b1;
				end
				else begin
					state <= waitForAvailableNode;
				end*/
				//state <= waitForAvailableNode;
				/*
				nodeWanted <= 4'd15;
				case(packet_in[79:74])
					6'd6: begin //FRAME SWITCH
						state <= sendCRCResultBack;
						readFrame <= ~readFrame;
						writeFrame <= ~writeFrame;
					end
					default: begin
						state <= waitForAvailableNode;
					end
				endcase*/
			end
			
			waitForAvailableNode: begin
				/*if(~packet_crc_in) begin
					state <= sendCRCResultBack;
					transmit_clock <= 1'b1;
				end
				else*/ begin
					nodeWanted <= oinstr[3:0];
					if(nodeWanted_isAvailable) begin
						state <= sendCommand;
					end
				end
			end
			
			sendCommand: begin
				command_out <= 1'b1;
				state <= waitForNodeActive;
			end
			
			waitForNodeActive: begin
				if(~nodeWanted_isAvailable) begin
					state <= waitForNodeComplete;
					command_out <= 1'b0;
				end
			end
			
			waitForNodeComplete: begin
				if(nodeWanted_isAvailable) begin
					state <= packetNotHere;
					transmit_clock <= 1'b1;
				end
			end
			
			packetNotHere: begin	
				if(packet_available == 1'b0) begin
					state <= sendCRCResultBack;
				end
				writeToIRAM <= 1'b0;
			end
			
			sendCRCResultBack: begin //
				
				//if(isTransmitting) begin
					//state <= waitForPacket;
					//transmit_clock <= 1'b0;
				//end
				//else begin
					//transmit_clock <= 1'b1;
					state <= waitForPacket; 
					transmit_clock <= 1'b0;
							/*
					bugCounter <= bugCounter + 1;
					if(bugCounter < 8'd80) begin
						transmit_clock <= 1'b0; end
						else if(bugCounter < 8'd160) begin
							transmit_clock <= 1'b1; end
							else begin transmit_clock <= 1'b0; end*/
				//end
			end
			
			default: begin
				state <= waitForPacket;
			end
		endcase
		end //if(instructionType == direct)
		else if(instructionType == download) begin
			case(download_state)
				waitForDownloadPacket: begin
					transmit_clock <= 1'b0;
					if(packet_available) begin
						download_state <= writePacket;
						grab_out <= 1'b1;
						writeToIRAM <= 1'b1;
					end
				end
				writePacket: begin
					if(packet_in[79:74] != 6'd15) begin
						
					end
					else begin
						
					end
					download_state <= sendACK;
					grab_out <= 1'b0;
				end
				sendACK: begin
					transmit_clock <= 1'b1;
					download_state <= waitTwo;
					writeToIRAM <= 1'b0;
				end
				waitTwo: begin
					if(packet_available == 1'b0) begin
						download_state <= waitForDownloadPacket;
					end
					transmit_clock <= 1'b0;
				end
				enter: begin
					transmit_clock <= 1'b1;
					grab_out <= 1'b0;
					if(packet_available == 1'b1) begin
						download_state <= waitForDownloadPacket;
					end
					else begin
						download_state <= waitTwo;
					end
					//download_state <= waitForDownloadPacket;
				end
				default: begin
					download_state <= waitForDownloadPacket;
				end
			endcase
		end
		else if(instructionType == memory) begin
			transmit_clock <= 1'b0;
			case(processor_state)
				waitInstruction: begin
					if(isDrawInstruction) begin
						processor_state <= parseInstruction;
						drawComplete <= 0;
					end
				end
				
				parseInstruction: begin
					oinstr <= IRAM_instrOut[79:74];
					ocolor <= IRAM_instrOut[73:62];
					oxstart <= IRAM_instrOut[61:52]; //x_start
					oystart <= IRAM_instrOut[51:43]; //y_start
					oxend_width <= IRAM_instrOut[42:33]; //x_end/width
					oyend_height <= IRAM_instrOut[32:24]; //y_end/height
					oextra1 <= IRAM_instrOut[23:16]; //extra_1
					oextra2 <= IRAM_instrOut[15:8]; //extra_2
					if(IRAM_instrOut[79:74] == 6'd6) begin//frame update
						processor_state <= packetNotHere;
						readFrame <= ~readFrame;
						writeFrame <= ~writeFrame;
					end
					else begin
						processor_state <= selectNode;
					end
				end
				
				selectNode: begin
					nodeWanted <= oinstr[3:0];
					processor_state <= issueCommand;
				end
				
				issueCommand: begin
					command_out <= 1'b1;
					processor_state <= waitForNodeToStart;
				end
				
				waitForNodeToStart: begin
					if(~nodeWanted_isAvailable) begin
						processor_state <= waitForNodeToFinish;
						command_out <= 1'b0;
					end
				end
				
				waitForNodeToFinish: begin
					if(nodeWanted_isAvailable) begin
						processor_state <= cleanUp;
						drawComplete <= 1;
					end
				end
				
				cleanUp: begin
					processor_state <= waitInstruction;
				end
			endcase
		end
		
	//end//end if(sync)
	end //end else statement

end





endmodule

